-- Initialise the resets array
rawset(_G, "kr_resets", {})

function A_KRCheckExistence(p)
	if p
	and p.valid
	and p.mo
	and p.mo.valid
		return true
	else
		return false
	end
end

function A_KRPlayerInit(p)
	p.krework = {}
	p.krework.pbuttons = p.cmd.buttons
	p.krework.jumpused = false
	p.krework.shieldt = 0
	p.krework.shieldused = false
	if p.krework.freeaim == nil
		p.krework.freeaim = 1
	end
	if p.krework.uncurltoggle == nil
		p.krework.uncurltoggle = true
	end
	if p.krework.superpowers == nil
		p.krework.superpowers = true
	end
	
	p.krework.inputAngle = function() return R_PointToAngle2(0, 0, p.cmd.forwardmove*FRACUNIT, -p.cmd.sidemove*FRACUNIT) end
	p.krework.realInputAngle = function() return p.krework.inputAngle() + p.cmd.angleturn<<16 end
	p.krework.moveAngle = function() return p.mo.angle + ((p.pflags & PF_ANALOGMODE) and 0 or p.krework.inputAngle()) end
	
	A_KRResetAbilities(p)
	return
end

-- Resets a player's abilities.
-- bop is true when this was triggered by hitting a monitor or an enemy
function A_KRResetAbilities(p, bop)
	local mo = p.mo
	
	if not mo.krework
		mo.krework = {}
	end
	
	-- Reset whether you've curled up mid-air with spin
	-- (Currently only used to eat Tails's remaining fly boosts when you aircurl)
	mo.krework.aircurled = false
	
	-- Reset whether you've used your jump ability
	p.krework.jumpused = false
	
	-- Reset whether you've used your shield ability
	if not (bop and (p.powers[pw_shield] & SH_NOSTACK) == SH_ATTRACT) -- Don't want attract shield to be chainable
		p.krework.shieldused = false
	end
	
	-- Reset the secondjump var
	p.secondjump = 0
	
	-- Reset the 'used ability' flags
	p.pflags = $ & ~PF_THOKKED & ~PF_SHIELDABILITY
	
	-- Resets for specific characters
	for skin, reset in pairs(kr_resets)
		if mo.skin == skin
			mo.krework[skin] = $ or {}
			kr_resets[skin](mo)
		else
			mo.krework[skin] = {}
		end
	end
end

-- Checks if two things overlap vertically
function A_KRCheckCollisionHeight(a, b)
	if not (a.valid and b.valid)
		return false
	end
	
	if a.z + a.momz + a.height < b.z + b.momz
	or b.z + b.momz + b.height < a.z + a.momz 
		return false
	end
	return true
end

-- Checks if you don't have any ability to control yourself
function A_KRCheckNoControl(p)
	if (P_PlayerInPain(p)
	or p.state == S_PLAY_PAIN
	or (p.playerstate == PST_DEAD)
	or p.powers[pw_carry])
		return true
	else
		return false
	end
end

-- Checks if you're one of the game's default characters, so we don't go too crazy with you
function A_KRCheckStockChar(p)
	if p.mo.skin == "sonic"
	or p.mo.skin == "tails"
	or p.mo.skin == "knuckles"
	or p.mo.skin == "amy"
	or p.mo.skin == "fang"
	or p.mo.skin == "metalsonic"
		return true
	else
		return false
	end
end

function A_KRAirjump(player)
	player.mo.state = S_PLAY_JUMP
	player.pflags = $ | PF_JUMPED
	
	if player.charability2 == CA2_SPINDASH
		player.pflags = $ & ~PF_NOJUMPDAMAGE
	end
end

function A_KRAircurl(p)
	p.mo.state = S_PLAY_JUMP
	p.mo.krework.aircurled = true
	p.pflags = $ | PF_JUMPED | PF_SPINNING & ~(PF_GLIDING) & ~(PF_THOKKED) & ~(PF_NOJUMPDAMAGE)
	S_StartSound(p.mo, sfx_curl)
	return true
end

function A_KRUncurl(p)
	local mo = p.mo
	
	-- Stop spinning
	p.pflags = $ & ~PF_SPINNING
	
	-- Walk or run
	if FixedHypot(mo.momx, mo.momy) >= p.runspeed
		mo.state = S_PLAY_RUN
	else
		mo.state = S_PLAY_WALK
	end
end