-- SETTINGS
-- Number of frames you must stay curled before you can uncurl
local UNCURL_LOCKTIME = 8
-- Reduction factor on boss bouncing
local BOSS_MOMENTUM_PERCENT = 75
-- Number of frames you have to let go of mid-air spin to use a shield power
local SHIELD_WINDOW = 7
-- Whether taking damage resets abilities
local PAIN_RESETS = false
-- Minimum spindash speed normally
local MIN_DASH = 20*FRACUNIT
-- Minimum spindash speed with super sneakers
local MIN_DASH_SUPER = 40*FRACUNIT

-- Handle pre-frame setup
addHook("PreThinkFrame", function()
	for player in players.iterate()
		if not A_KRCheckExistence(player)
			continue
		end
		
		-- Init
		if not player.krework
			A_KRPlayerInit(player)
		end

		if not player.mo.krework
			A_KRResetAbilities(player)
		end
		
		local mo = player.mo
		
		-- Dodgy flags, don't use if we can avoid
		if not ((player.powers[pw_shield] & SH_NOSTACK) == SH_BUBBLEWRAP
			or (player.powers[pw_shield] & SH_NOSTACK) == SH_ELEMENTAL
			or (player.powers[pw_shield] & SH_NOSTACK) == SH_FLAMEAURA)
			player.pflags = $ & ~PF_SHIELDABILITY
		end
		player.pflags = $ & ~PF_THOKKED
		
		/*
		if mo.state == S_PLAY_CLING or mo.state == S_PLAY_CLIMB
			// Force camera centering when I can figure out how
		end
		*/
		
		-- Spindash buffs
		if player.charability2 == CA2_SPINDASH
			-- Min speed buff
			-- Sneaker-buffed mindash
			if player.powers[pw_sneakers] or player.powers[pw_super]
				if not player.krework.mindash
					player.krework.mindash = player.mindash
				end
				player.mindash = max($, MIN_DASH_SUPER)
			-- Sneakers are over, so restore their old mindash
			elseif player.krework.mindash
				player.mindash = player.krework.mindash
				player.krework.mindash = nil
			-- Regular buffed mindash
			else
				player.mindash = max($, MIN_DASH)
			end
			
			-- Autospindash on slow landing
			if (mo.eflags & MFE_JUSTHITFLOOR) and (player.cmd.buttons & BT_USE)
				player.pflags = $ & ~PF_USEDOWN
			end
		else
			player.mindash = skins[player.mo.skin].mindash
		end
		
		-- If you have the whirlwind shield and you've used its ability and your jump ability, no jump 4 u
		if (player.powers[pw_shield] & SH_NOSTACK) == SH_WHIRLWIND
		and player.krework.jumpused and player.krework.shieldused 
			player.cmd.buttons = $ & ~BT_JUMP
		end
		
		-- Enter jump state mid-air to enable air abilities
		if not P_IsObjectOnGround(mo) and mo.state != S_PLAY_JUMP -- You're in the air but not because you jumped to get there
		and mo.state != S_PLAY_CLING and mo.state != S_PLAY_CLIMB -- You're not climbing walls
		and mo.state != S_PLAY_GASP -- You're not currently sucking down a bubble
		and (player.cmd.buttons & BT_JUMP) and not (player.pflags & PF_JUMPDOWN) -- You've just pressed jump
		and not A_KRCheckNoControl(player) -- You have control
		and not player.powers[pw_carry] -- You're not being carried by something
		and not player.krework.jumpused -- You haven't fully used your ability
		and not (mo.skin == "tails" and player.bot) -- You're not a bot Tails
			A_KRAirjump(player)
		end
		
		-- Decide whether it's time to use a shield ability
		player.krework.shieldready = false
		if not (player.cmd.buttons & BT_USE)
		and (player.krework.pbuttons & BT_USE)	
		and player.krework.shieldt > 0
		and player.krework.shieldt <= SHIELD_WINDOW
			player.krework.shieldready = true
			player.cmd.buttons = $ | BT_USE
			player.pflags = $ & ~PF_USEDOWN
		end
		
		-- Increment the shield counter
		if not A_KRCheckNoControl(player)
		and (player.cmd.buttons & BT_USE)
		and (player.pflags & PF_JUMPED)
			player.krework.shieldt = $+1
		end
		
		-- Reset air abilities:
		if (P_IsObjectOnGround(mo) and -- On being legitimately on the ground
			(mo.state == S_PLAY_STND
			or mo.state == S_PLAY_WAIT
			or mo.state == S_PLAY_WALK
			or mo.state == S_PLAY_SKID
			or mo.state == S_PLAY_RUN
			or mo.state == S_PLAY_DASH
			or mo.state == S_PLAY_ROLL
			or mo.state == S_PLAY_EDGE
			or mo.state == S_PLAY_PAIN
			or mo.state == S_PLAY_STUN
			or mo.state == S_PLAY_GASP))
		or (mo.eflags & MFE_SPRUNG) -- On springing
		or mo.state == S_PLAY_CLING or mo.state == S_PLAY_CLIMB -- On climbing a wall
		or player.secondjump == UINT8_MAX -- On Bubble Shield bouncing
		or player.powers[pw_carry] -- On getting carried by something
		or (mo.momz == (FRACUNIT * 39 / 2) and mo.state == S_PLAY_FALL) -- On lightly brushing the top of a Dust Devil // VERY HACKY
		or (PAIN_RESETS and A_KRCheckNoControl(player)) -- When you get hit
			A_KRResetAbilities(player)
		end
	end
end)

-- Generic ability condition replacing the function of PF_THOKKED
addHook("AbilitySpecial", function(player)
	-- Check that you haven't used your jump ability yet
	if player.krework and player.krework.jumpused
		player.pflags = $ | PF_THOKKED
		return true
	end
	
	
	return false
end)

-- Toggled uncurling and aircurling for all characters
addHook("SpinSpecial", function(player)
	-- Have to exist to uncurl
	if not A_KRCheckExistence(player)
		return false
	end
	
	local mo = player.mo
	
	-- On ground, attempt toggled uncurl
	if P_IsObjectOnGround(mo)
		if mo.state == S_PLAY_ROLL -- Must be curled
		and player.krework.uncurltoggle -- Uncurl toggle must be on
		and not A_KRCheckNoControl(player) -- Control must be available
		and not mo.krework.uncurl_lock -- Uncurling can't be locked
		and not (player.krework.pbuttons & BT_USE) -- Can't be already holding spin
			A_KRUncurl(player)
			return true
		end
	else
		-- Try to curl in the air
		if player.charability2 == CA2_SPINDASH -- Must be able to spin
		and mo.state != S_PLAY_ROLL -- Can't already be curled up
		and mo.state != S_PLAY_JUMP -- Can't already be curled up
		and mo.state != S_PLAY_CLING -- Wall clinging is not the air
		and mo.state != S_PLAY_CLIMB -- Wall climbing is not the air
		and not (player.krework.pbuttons & BT_USE) -- Must be pressing spin, not holding it
		and not (player.powers[pw_carry]) -- Can't be being carried by something
			A_KRAircurl(player)
			return true
		end
	end
end)

addHook("ShieldSpecial", function(player)
	-- Don't use shield if it's not ready
	if not player.krework.shieldready
	-- Bubble shield can be used repeatedly, all other shields can't
	or player.krework.shieldused and (player.powers[pw_shield] & SH_NOSTACK) != SH_BUBBLEWRAP
		return true
	end
	
	-- These shields need PF_SHIELDABILITY for their longer-lasting effects
	if (player.powers[pw_shield] & SH_NOSTACK) == SH_BUBBLEWRAP -- Bounces off the ground
	or (player.powers[pw_shield] & SH_NOSTACK) == SH_ELEMENTAL -- The firey splat after landing
	or (player.powers[pw_shield] & SH_NOSTACK) == SH_FLAMEAURA -- Keeps burning until you land
		player.pflags = $ | PF_SHIELDABILITY
	end
	
	player.krework.shieldused = true
end)

-- Various controls for all characters
addHook("PlayerThink", function(player)
	-- Big list of reasons to bail out early
	if not A_KRCheckExistence(player)
	or not (player.playerstate == PST_LIVE)
	or player.exiting
	or player.powers[pw_nocontrol]
	or P_PlayerInPain(player)
		return
	end
	
	local mo = player.mo
	
	-- General moveset tweaks
		
	-- Untoggled uncurling
	if not player.krework.uncurltoggle -- Uncurling is on release, not toggle
	and P_IsObjectOnGround(mo) -- You're on the ground
	and mo.state == S_PLAY_ROLL -- You're curled
	and not A_KRCheckNoControl(player) -- Control is available
	and not mo.krework.uncurl_lock -- Uncurling is not locked
	and not (player.cmd.buttons & BT_USE) -- You're not holding spin
		A_KRUncurl(player)
	end
	
	-- Uncurl lock timer
	-- If you're rolling, count down the timer
	if P_IsObjectOnGround(mo) and mo.state == S_PLAY_ROLL 
		if mo.krework.uncurl_lock
			mo.krework.uncurl_lock = max(0, $ - 1)
		end
	-- If you're not, reset the timer
	else
		mo.krework.uncurl_lock = UNCURL_LOCKTIME
	end
	
	-- Full air control when jumping out of a roll for all characters
	if player.pflags & PF_SPINNING and not P_IsObjectOnGround(mo)
		if player.pflags & PF_JUMPED
			player.pflags = $ & ~PF_SPINNING
		end
	end
end)

-- Refresh stock char abilities after hitting monitor/enemy/boss
addHook("MobjMoveCollide", function(mo, touched)
	local player = mo.player
	
	if not A_KRCheckExistence(player)
	or not A_KRCheckCollisionHeight(mo, touched)
		return
	end
	
	if not (touched.flags & (MF_MONITOR|MF_ENEMY|MF_BOSS))
	or not (touched.flags & MF_SHOOTABLE)
	or (touched.flags & MF_MISSILE)
	or not P_PlayerCanDamage(player, touched)
		return
	end
	
	A_KRResetAbilities(player, true)
		
	-- Reduce x and y velocity when bouncing off a boss/facestabber/etc
	if (touched.flags & MF_BOSS) or (touched.health > 1)
		mo.momx = $ * BOSS_MOMENTUM_PERCENT / 100
		mo.momy = $ * BOSS_MOMENTUM_PERCENT / 100
	end
end, MT_PLAYER)

-- Cleanup
addHook("PostThinkFrame", function()
	for player in players.iterate()
		-- Existence check
		if not A_KRCheckExistence(player)
			continue
		end
		
		local mo = player.mo
		
		-- Cut the aircurl sound early if you're not actually curled
		if mo.krework.aircurled and not (mo.state == S_PLAY_JUMP or mo.state == S_PLAY_ROLL)
			S_StopSoundByID(mo, sfx_curl)
		end
		
		-- Record this frame's buttons for next frame
		player.krework.pbuttons = player.cmd.buttons
		
		-- Reset the aircurled check
		mo.krework.aircurled = false
		
		-- Reset the shield ability timing counter
		if not (player.cmd.buttons & BT_USE)
		or P_IsObjectOnGround(mo)
		or A_KRCheckNoControl(player)
		or player.krework.shieldready
			player.krework.shieldt = 0
		end
		
		-- Our substitute thokked flag because the real one has too much baggage attached
		-- Flyers don't use the thokked flag properly
		if player.charability == CA_FLY
			if mo.state == S_PLAY_FLY_TIRED
				player.krework.jumpused = true
			end
		-- Everyone else does, except that shield abilities also set thokked
		else
			if (player.pflags & PF_THOKKED) and not player.krework.shieldready
				player.krework.jumpused = true
			end
		end
	end
end)
