-- SETTINGS
-- Sonic's drop dash
local DROP_LV1TIME = 10
local DROP_LV2TIME = 45
local DROP_LV3TIME = 85

local DROP_LV1 = {level = 1, speed = 40*FRACUNIT, dustamt = 4}
local DROP_LV2 = {level = 2, speed = 60*FRACUNIT, dustamt = 6}
local DROP_LV3 = {level = 3, speed = 80*FRACUNIT, dustamt = 8}

-- In simple strafe mode, setting this to true makes L/R input act like a spindash and rotate you while charging, not move
local DROP_ROTAIM = false

-- Alternate color for level 3 dropdash trail
local TRAIL_COLORED = false
local TRAIL_COLOR = SKINCOLOR_CYAN

-- Sonic's double jump
local DJ_SPEED = 8*FRACUNIT
local DJ_MOMZ = (19/2)*FRACUNIT

rawset(_G["kr_resets"], "sonic", function(mo)
	-- Reset how many frames Sonic has charged drop dash
	mo.krework.sonic.dropcharge = (type($) == "number") and $ or 0
	-- Initialise whether Sonic is charging this frame, just to be safe
	mo.krework.sonic.charging = $ and true or false
end)

-- Sonic double jump
function A_KRSonicDoubleJump(player)
	if not A_KRCheckExistence(player)
	or not (player.mo.skin == "sonic")
	or not (player.pflags & PF_JUMPED)
	or (not player.krework.superpowers and player.powers[pw_super])
		return false
	end
	
	local mo = player.mo
	
	if player.krework.jumpused
		return true
	end
	
	S_StartSound(mo, sfx_dbjump)
	
	-- Initialise vars
	local speed = FixedHypot(mo.momx, mo.momy)
	local momangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local djspeed = DJ_SPEED
	local djmomz = DJ_MOMZ
	local dustamt = 4
	local abilityaim = player.krework.realInputAngle()
	
	-- Depending on the sharpness of your turn, you will lose speed
	local anglediff = AngleFixed(abs(momangle - player.krework.realInputAngle())) / FRACUNIT
	local speedpercentage = max(0, min(100, (100 - (anglediff - 50))))
	speed = $ * speedpercentage / 100
	
	-- Flags, states
	mo.krework.sonic.dropcharge = 0 -- Reset drop dash timer
	player.pflags = $ | PF_JUMPED
	player.pflags = $ | PF_STARTJUMP
	mo.state = S_PLAY_JUMP
	
	-- Underwater and super speed modifiers
	if mo.eflags&MFE_UNDERWATER
		djspeed = $*2/3
		djmomz = $*2/3
	end
	if player.powers[pw_super]
		djspeed = $*3/2
		djmomz = $*3/2
		local dustamt = 6
	end
	
	-- Apply vertical thrust
	P_SetObjectMomZ(mo, djmomz, false)
	
	-- Apply horizontal thrust unless the movement controls are mostly neutral
	if FixedHypot(abs(player.cmd.forwardmove), abs(player.cmd.sidemove)) > 5
		P_InstaThrust(mo, abilityaim, max(djspeed, speed))
		if (player.pflags & PF_ANALOGMODE) or (player.pflags & PF_DIRECTIONCHAR)
			player.drawangle = abilityaim
		end
	end
	
	-- Dust
	local function r(mo, value)
		local w = mo.radius/FRACUNIT/2
		return value + P_RandomRange(-w, w) * FRACUNIT
	end
	
	for n = 1, dustamt
		local d = P_SpawnMobj(r(mo, mo.x), r(mo, mo.y), mo.z, MT_SPINDUST)
		d.extravalue1 = FRACUNIT
		d.extravalue2 = FRACUNIT
		if P_MobjFlip(mo) == -1
			d.eflags = $|MFE_VERTICALFLIP
		end
		P_SetObjectMomZ(d, FRACUNIT*2)
		local l = 180*n/dustamt
		P_InstaThrust(d, player.drawangle+ANGLE_135+l*ANG1, DJ_SPEED/2)
	end
	
	player.krework.jumpused = true
	return true
end

function A_KRChargeDropDash(player)
	local mo = player.mo
	local charge = mo.krework.sonic.dropcharge
	
	-- Increase dropdash charge, cap at level 3 + 1 frame
	local function chargeUp()
		charge = min($ + 1, DROP_LV3TIME + 1)
	
		-- SFX for reaching a new charge level
		if charge == DROP_LV1TIME
			S_StartSound(mo, sfx_drpdsh)
		elseif charge == DROP_LV2TIME
			S_StartSound(mo, sfx_s3kab8)
		elseif charge == DROP_LV3TIME
			S_StartSound(mo, sfx_s3kabf)
		end
	end
	
	chargeUp()
	-- Charge twice as fast with sneakers
	if player.powers[pw_sneakers] or player.powers[pw_super]
		chargeUp()
	end
	
	mo.krework.sonic.dropcharge = charge
	
	-- Dropdashing is LIKE spindashing
	if DROP_ROTAIM and not (player.pflags & PF_ANALOGMODE)
		player.pflags = $ | PF_STARTDASH
	else
		player.pflags = $ & ~PF_STARTDASH
	end
	
	-- Ultra drop dash is ready
	if charge >= DROP_LV3TIME
		local trail = P_SpawnGhostMobj(mo)
		P_TeleportMove(trail, trail.x, trail.y, trail.z - 7 * FRACUNIT)
		trail.fuse = 12
		trail.state = S_THOK
		trail.scale = FRACUNIT
		trail.destscale = FRACUNIT/4
		if TRAIL_COLORED
			trail.color = TRAIL_COLOR
			trail.colorized = true
		end
		
	-- No? Then super drop dash is ready
	elseif charge >= DROP_LV2TIME
		local trail = P_SpawnGhostMobj(mo)
		P_TeleportMove(trail, trail.x, trail.y, trail.z - 3 * FRACUNIT)
		trail.fuse = 8
		trail.state = S_THOK
		trail.scale = FRACUNIT * 4/5 
		trail.destscale = FRACUNIT/4
		trail.frame = TR_TRANS80|A
		
	-- No? Then regular drop dash is ready
	elseif charge >= DROP_LV1TIME
		if mo.state == S_PLAY_JUMP
			mo.state = S_DROPDASH1
		end
	end
	
	player.mo.krework.sonic.charging = true
	
	-- Always face forwards when charging dropdash
	if not player.krework.freeaim
		player.drawangle = mo.angle
	end
end

function A_KRReleaseDropDash(player)
	local mo = player.mo
	local charge = mo.krework.sonic.dropcharge
	
	if charge < DROP_LV1TIME
		return
	end
	
	local speed = FixedHypot(mo.momx, mo.momy)
	local dropdata = DROP_LV1
	
	-- Decide which way to fire the dropdash
	-- Simple mode or freeaim directionchar
	if player.pflags & PF_ANALOGMODE
	or (player.pflags & PF_DIRECTIONCHAR and player.krework.freeaim)
		-- Go in the direction you're holding on the stick
		player.drawangle = player.krework.realInputAngle()
	-- Strafe mode or no freeaim directionchar
	else
		-- Just go forward
		player.drawangle = mo.angle
	end
	
	-- Decide which charge level you're at
	if charge >= DROP_LV2TIME
		if charge < DROP_LV3TIME
			dropdata = DROP_LV2
		else
			dropdata = DROP_LV3
		end
	end
	
	-- State change
	mo.state = S_PLAY_ROLL
	player.pflags = $1|PF_SPINNING
	
	-- Thrust
	local dropspeed = dropdata.speed
		-- Underwater and super speed modifiers
		if mo.eflags&MFE_UNDERWATER
			dropspeed = $*2/3
		end
		if player.powers[pw_super]
			dropspeed = $*3/2
		end
	
	P_InstaThrust(mo, player.drawangle, max(dropspeed, speed))
	
	-- Sounds
	S_StartSound(mo, sfx_zoom)
	if dropdata.level >= 2
		S_StartSound(mo, sfx_s3k81)
	end
	if dropdata.level == 3
		S_StartSound(mo, sfx_s3ka2)
	end
	
	-- Dust
	local dustamt = dropdata.dustamt
	
	local function r(mo, value)
		local w = mo.radius/FRACUNIT/2
		return value + P_RandomRange(-w, w) * FRACUNIT
	end
	
	for n = 1, dustamt
		local d = P_SpawnMobj(r(mo, mo.x), r(mo, mo.y), mo.z, MT_SPINDUST)
		d.extravalue1 = FRACUNIT
		d.extravalue2 = FRACUNIT
		if P_MobjFlip(mo) == -1
			d.eflags = $|MFE_VERTICALFLIP
		end
		P_SetObjectMomZ(d, FRACUNIT*2)
		local l = 90*n/dustamt
		P_InstaThrust(d, player.drawangle+ANGLE_135+l*ANG1, dropspeed/2)
	end
	
	-- Quake
	if dropdata.level == 3
		P_StartQuake(30*FRACUNIT, 5, {mo.x, mo.y, mo.z}, 200 * FRACUNIT)
	end
	
	mo.krework.sonic.dropcharge = 0
end
