-- SETTINGS
-- Number of times you can boost per jump
local FLY_COUNT = 3
-- Boost ascent speed
local FLY_ASCENDSPEED = (53/20)*FRACUNIT
-- Number of tics you spend boosting
local FLY_ASCENDTICS = 7
-- Number of tics you spend boosting when swimming
local FLY_ASCENDTICS_SWIM = 4
-- Minimum time between flight boosts
local FLY_SPAMTICS = 10
-- How fast you fall when first hovering
local FLY_HOVERSPEED = -(7/2) * FRACUNIT
-- How fast you fall towards the end of your hover
local FLY_HOVERSPEED_TIRED = -(11/2) * FRACUNIT
-- How many tics it takes to hit the end of hover state
local FLY_TIREDTIME = 55
-- How many tics it takes to drop out of the air entirely
local FLY_MAXTIME = 85
-- Whether to resize the dust clouds produced by flight boosts
local DUST_SCALE = true
-- Whether to color the dust clouds produced by flight boosts
local DUST_COLOR = false
-- Colors for each stage of dust
local DUST_COLORS = {[2] = SKINCOLOR_YELLOW, [1] = SKINCOLOR_ORANGE, [0] = SKINCOLOR_RED}
local DUST_DEFAULT = SKINCOLOR_EMERALD

rawset(_G["kr_resets"], "tails", function(mo)
    -- Reset how long Tails has been flying
    mo.krework.tails.flytime = 0
    -- Reset how many more times Tails can boost
    A_KRTailsResetFlyCount(mo.player)
end)

function A_KRTailsResetFlyCount(p)
	p.mo.krework.tails.flycount = FLY_COUNT
	p.abilityused = false
end

-- Tails flight activated
function A_KRTailsFly(player)
	if not A_KRCheckExistence(player)
	or not (player.mo.skin == "tails")
	or player.bot == 1
	or not (player.pflags & PF_JUMPED)
		return false
	end
	
	local mo = player.mo
	local speed = FixedHypot(mo.momx, mo.momy)
	local momangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	local time = mo.krework.tails.flytime
	
	if player.krework.jumpused
	or (time > 0 and time < FLY_SPAMTICS)
		return true
	end
	
	-- Momentum
	mo.momz = $ / 2
	P_Thrust(mo, momangle + ANGLE_180, speed / 4)
	
	-- Flags, states
	mo.krework.aircurled = false
	mo.krework.tails.flytime = 0
	mo.krework.tails.flycount = (mo.eflags & MFE_UNDERWATER) and FLY_COUNT or ($1 - 1)
	mo.krework.tails.flylock = FLY_SPAMTICS
	player.pflags = $ | PF_NOJUMPDAMAGE
	player.pflags = $ & ~PF_SPINNING
	mo.state = S_PLAY_FLY
	
	-- Sound effect
	if not (mo.eflags & MFE_UNDERWATER)
		S_StartSound(mo, sfx_putput)
	end
	return true
end

function A_KRTailsHover(player)
	local mo = player.mo
	local time = mo.krework.tails.flytime

	-- Declare flight-related variables
	local hoverspeed = FLY_HOVERSPEED
	local flip = P_MobjFlip(mo)
	
	-- Change state to tired to show when you can boost again
	if time >= FLY_SPAMTICS
		if mo.state == S_PLAY_FLY
			mo.state = S_PLAY_FLY_TIRED
		end
	else
		player.krework.jumpused = true
	end
	
	-- Change sound when you hit tired physics
	if time < FLY_TIREDTIME
		if time % 10 == 1
		and (mo.state == S_PLAY_FLY or mo.state == S_PLAY_FLY_TIRED)
			S_StartSound(mo, sfx_s3kba)
		end
	else
		if mo.state == S_PLAY_FLY_TIRED 
			if mo.eflags & MFE_UNDERWATER
				mo.state = S_PLAY_FLY
			elseif time % 5 == 1
				S_StartSound(mo, sfx_s3kbb)
			end
		end
		
		-- Start descending slightly faster if in the air
		if not (mo.eflags & MFE_UNDERWATER)
			hoverspeed = FLY_HOVERSPEED_TIRED
		end
	end
	
	-- Decide on vertboost duration
	local ascendtics = FLY_ASCENDTICS
	if mo.eflags & MFE_UNDERWATER
		ascendtics = FLY_ASCENDTICS_SWIM
	end
	local speed = FixedHypot(mo.momx, mo.momy)
	
	//print(speed/FRACUNIT)
	
	-- Fall no faster than your hover speed
	if (flip == 1 and mo.momz < hoverspeed) or (flip == -1 and mo.momz > -hoverspeed)
		P_SetObjectMomZ(mo, hoverspeed, false)
	end
	//P_SetObjectMomZ(mo, min(flippedmomz, hoverspeed))
	
	-- If you're still in vertboost, apply boost
	if (time <= ascendtics)
		-- Calculate boost to apply
		local ascendspeed = FLY_ASCENDSPEED
		if mo.eflags&MFE_UNDERWATER
			ascendspeed = $ * 2/3
		end
		
		P_SetObjectMomZ(mo, ascendspeed, true)
		
		-- Dust clouds (or bubbles) to represent Tails boosting
		if time % 2 == 0
			local source
			local dusttype
			local dust1x
			local dust1y
			local dust2x
			local dust2y
			local dustz
			local dustscale
			local dustcolor
			
			-- Spawn bubbles behind Tails's feet when swimming
			if mo.eflags&MFE_UNDERWATER
				source = player.mo
				
				-- Random deviation from center
				local bubblewobble = 2 * FRACUNIT / 5
				-- Need this because P_RandomRange can't handle a range >=65536
				local bubblewobblemult = 2
				
				local angle = (source.angle + ANGLE_157h)
				dust1x = FixedMul(cos(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust1y = FixedMul(sin(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust2x = FixedMul(cos(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust2y = FixedMul(sin(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dustz = 0
				
				dusttype = MT_MEDIUMBUBBLE
				
				dustscale = false
				dustcolor = false
			-- Spawn dust clouds above Tails's tails when flying
			else
				source = player.followmobj
				
				-- Tails's tails have a 4-frame animation that rotates 180 degrees
				local angle = (player.mo.angle + (source.frame * ANGLE_45))
				dust1x = FixedMul(cos(angle), source.radius)
				dust1y = FixedMul(sin(angle), source.radius)
				dust2x = FixedMul(cos(angle+ANGLE_180), source.radius)
				dust2y = FixedMul(sin(angle+ANGLE_180), source.radius)
				dustz = source.height*4/5
				
				dusttype = MT_SPINDUST
				
				dustscale = DUST_SCALE
				dustcolor = DUST_COLOR
			end
			
			local d1 = P_SpawnMobjFromMobj(source, dust1x, dust1y, dustz, dusttype)
			local d2 = P_SpawnMobjFromMobj(source, dust2x, dust2y, dustz, dusttype)
			
			-- Change color to represent spending boosts
			if dustcolor and mo.krework.tails.flycount < 3
				local dcolor = DUST_COLORS[mo.krework.tails.flycount] or DUST_DEFAULT
				d1.color, d2.color = dcolor, dcolor
				d1.colorized, d2.colorized = true, true
			end
			
			-- Reduce in size to represent spending boosts
			if dustscale
				d1.scale = $ * (mo.krework.tails.flycount+1) / 3
				d2.scale = d1.scale
			end
		end
	end
	
	-- Drop out of the air, keep swimming in water
	if time >= FLY_MAXTIME and not (mo.eflags&MFE_UNDERWATER)
		mo.krework.tails.flytime = 0
		mo.state = S_PLAY_FALL
	else
		mo.krework.tails.flytime = $ + 1
	end
end
